@@ -1,21 +1,21 @@
 #define BOOST_TEST_DYN_LINK
 #define BOOST_TEST_NO_MAIN
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 #include <boost/test/unit_test.hpp>
 #include <boost/assign/list_of.hpp>
-#include <boost/foreach.hpp>
+
 #include <boost/tuple/tuple.hpp>
 #include "pdns/namespaces.hh"
-#include <pdns/dns.hh>
-#include <pdns/dnsbackend.hh>
-#include <pdns/dnspacket.hh>
-#include <pdns/ueberbackend.hh>
-#include <pdns/pdnsexception.hh>
-#include <pdns/logger.hh>
-#include <pdns/arguments.hh>
-#include <boost/lexical_cast.hpp>
-#include <rapidjson/rapidjson.h>
-#include <rapidjson/document.h>
+#include "pdns/dns.hh"
+#include "pdns/dnsbackend.hh"
+#include "pdns/dnspacket.hh"
+#include "pdns/ueberbackend.hh"
+#include "pdns/pdnsexception.hh"
+#include "pdns/logger.hh"
+#include "pdns/arguments.hh"
 #include "pdns/json.hh"
 #include "pdns/statbag.hh"
 #include "pdns/packetcache.hh"
@@ -29,20 +29,20 @@ BOOST_AUTO_TEST_SUITE(test_remotebackend_so)
 BOOST_AUTO_TEST_CASE(test_method_lookup) {
    BOOST_TEST_MESSAGE("Testing lookup method");
    DNSResourceRecord rr;
-   be->lookup(QType(QType::SOA), "unit.test");
+   be->lookup(QType(QType::SOA), DNSName("unit.test."));
    // then try to get()
    BOOST_CHECK(be->get(rr)); // and this should be TRUE.
    // then we check rr contains what we expect
-   BOOST_CHECK_EQUAL(rr.qname, "unit.test");
+   BOOST_CHECK_EQUAL(rr.qname.toString(), "unit.test.");
    BOOST_CHECK_MESSAGE(rr.qtype == QType::SOA, "returned qtype was not SOA");
-   BOOST_CHECK_EQUAL(rr.content, "ns.unit.test hostmaster.unit.test 1 2 3 4 5 6");
+   BOOST_CHECK_EQUAL(rr.content, "ns.unit.test. hostmaster.unit.test. 1 2 3 4 5 6");
    BOOST_CHECK_EQUAL(rr.ttl, 300);
 }
 
 BOOST_AUTO_TEST_CASE(test_method_lookup_empty) {
    BOOST_TEST_MESSAGE("Testing lookup method with empty result");
    DNSResourceRecord rr;
-   be->lookup(QType(QType::SOA), "empty.unit.test");
+   be->lookup(QType(QType::SOA), DNSName("empty.unit.test."));
    // then try to get()
    BOOST_CHECK(!be->get(rr)); // and this should be FALSE
 }
@@ -52,7 +52,7 @@ BOOST_AUTO_TEST_CASE(test_method_list) {
    DNSResourceRecord rr;
 
    BOOST_TEST_MESSAGE("Testing list method");
-   be->list("unit.test", -1);
+   be->list(DNSName("unit.test."), -1);
    while(be->get(rr)) record_count++;
 
    BOOST_CHECK_EQUAL(record_count, 5); // number of records our test domain has
@@ -67,13 +67,13 @@ BOOST_AUTO_TEST_CASE(test_method_setDomainMetadata) {
    std::vector<std::string> meta;
    meta.push_back("VALUE");
    BOOST_TEST_MESSAGE("Testing setDomainMetadata method");
-   BOOST_CHECK(be->setDomainMetadata("unit.test","TEST", meta));
+   BOOST_CHECK(be->setDomainMetadata(DNSName("unit.test."),"TEST", meta));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getDomainMetadata) {
    std::vector<std::string> meta;
    BOOST_TEST_MESSAGE("Testing getDomainMetadata method");
-   be->getDomainMetadata("unit.test","TEST", meta);
+   be->getDomainMetadata(DNSName("unit.test."),"TEST", meta);
    BOOST_CHECK_EQUAL(meta.size(), 1);
    // in case we got more than one value, which would be unexpected
    // but not fatal
@@ -84,7 +84,7 @@ BOOST_AUTO_TEST_CASE(test_method_getDomainMetadata) {
 BOOST_AUTO_TEST_CASE(test_method_getAllDomainMetadata) {
    std::map<std::string, std::vector<std::string> > meta;
    BOOST_TEST_MESSAGE("Testing getAllDomainMetadata method");
-   be->getAllDomainMetadata("unit.test", meta);
+   be->getAllDomainMetadata(DNSName("unit.test."), meta);
    BOOST_CHECK_EQUAL(meta.size(), 1);
    // in case we got more than one value, which would be unexpected
    // but not fatal
@@ -94,21 +94,21 @@ BOOST_AUTO_TEST_CASE(test_method_getAllDomainMetadata) {
 
 BOOST_AUTO_TEST_CASE(test_method_addDomainKey) {
    BOOST_TEST_MESSAGE("Testing addDomainKey method");
-   BOOST_CHECK_EQUAL(be->addDomainKey("unit.test",k1), 1);
-   BOOST_CHECK_EQUAL(be->addDomainKey("unit.test",k2), 2);    
+   BOOST_CHECK_EQUAL(be->addDomainKey(DNSName("unit.test."),k1), 1);
+   BOOST_CHECK_EQUAL(be->addDomainKey(DNSName("unit.test."),k2), 2);
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getDomainKeys) {
    std::vector<DNSBackend::KeyData> keys;
    BOOST_TEST_MESSAGE("Testing getDomainKeys method");
    // we expect to get two keys
-   be->getDomainKeys("unit.test",0,keys);
+   be->getDomainKeys(DNSName("unit.test."),0,keys);
    BOOST_CHECK_EQUAL(keys.size(), 2);
    // in case we got more than 2 keys, which would be unexpected
    // but not fatal
    if (keys.size() > 1) {
       // check that we have two keys
-      BOOST_FOREACH(DNSBackend::KeyData &kd, keys) {
+      for(DNSBackend::KeyData &kd :  keys) {
         BOOST_CHECK(kd.id > 0);
         BOOST_CHECK(kd.flags == 256 || kd.flags == 257);
         BOOST_CHECK(kd.active == true);
@@ -119,47 +119,49 @@ BOOST_AUTO_TEST_CASE(test_method_getDomainKeys) {
 
 BOOST_AUTO_TEST_CASE(test_method_deactivateDomainKey) {
    BOOST_TEST_MESSAGE("Testing deactivateDomainKey method");
-   BOOST_CHECK(be->deactivateDomainKey("unit.test",1));
+   BOOST_CHECK(be->deactivateDomainKey(DNSName("unit.test."),1));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_activateDomainKey) {
    BOOST_TEST_MESSAGE("Testing activateDomainKey method");
-   BOOST_CHECK(be->activateDomainKey("unit.test",1));
+   BOOST_CHECK(be->activateDomainKey(DNSName("unit.test."),1));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_removeDomainKey) {
-   BOOST_CHECK(be->removeDomainKey("unit.test",2));
-   BOOST_CHECK(be->removeDomainKey("unit.test",1));
+   BOOST_CHECK(be->removeDomainKey(DNSName("unit.test."),2));
+   BOOST_CHECK(be->removeDomainKey(DNSName("unit.test."),1));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getBeforeAndAfterNamesAbsolute) {
-   std::string unhashed,before,after;
+   DNSName unhashed;
+   std::string before,after;
    BOOST_TEST_MESSAGE("Testing getBeforeAndAfterNamesAbsolute method");
    
-   be->getBeforeAndAfterNamesAbsolute(-1, "middle.unit.test", unhashed, before, after);
-   BOOST_CHECK_EQUAL(unhashed, "middle");
-   BOOST_CHECK_EQUAL(before, "begin");
-   BOOST_CHECK_EQUAL(after, "stop");
+   be->getBeforeAndAfterNamesAbsolute(-1, "middle.unit.test.", unhashed, before, after);
+   BOOST_CHECK_EQUAL(unhashed.toString(), "middle.");
+   BOOST_CHECK_EQUAL(before, "begin.");
+   BOOST_CHECK_EQUAL(after, "stop.");
 }
 
 BOOST_AUTO_TEST_CASE(test_method_setTSIGKey) {
    std::string algorithm, content;
    BOOST_TEST_MESSAGE("Testing setTSIGKey method");
-   BOOST_CHECK_MESSAGE(be->setTSIGKey("unit.test","hmac-md5","kp4/24gyYsEzbuTVJRUMoqGFmN3LYgVDzJ/3oRSP7ys="), "did not return true");
+   BOOST_CHECK_MESSAGE(be->setTSIGKey(DNSName("unit.test."),DNSName("hmac-md5."),"kp4/24gyYsEzbuTVJRUMoqGFmN3LYgVDzJ/3oRSP7ys="), "did not return true");
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getTSIGKey) {
-   std::string algorithm, content;
+   DNSName algorithm;
+   std::string content;
    BOOST_TEST_MESSAGE("Testing getTSIGKey method");
-   be->getTSIGKey("unit.test",&algorithm,&content);
-   BOOST_CHECK_EQUAL(algorithm, "hmac-md5");
+   be->getTSIGKey(DNSName("unit.test."),&algorithm,&content);
+   BOOST_CHECK_EQUAL(algorithm.toString(), "hmac-md5.");
    BOOST_CHECK_EQUAL(content, "kp4/24gyYsEzbuTVJRUMoqGFmN3LYgVDzJ/3oRSP7ys=");
 }
 
 BOOST_AUTO_TEST_CASE(test_method_deleteTSIGKey) {
    std::string algorithm, content;
    BOOST_TEST_MESSAGE("Testing deleteTSIGKey method");
-   BOOST_CHECK_MESSAGE(be->deleteTSIGKey("unit.test"), "did not return true");
+   BOOST_CHECK_MESSAGE(be->deleteTSIGKey(DNSName("unit.test.")), "did not return true");
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getTSIGKeys) {
@@ -168,8 +170,8 @@ BOOST_AUTO_TEST_CASE(test_method_getTSIGKeys) {
    be->getTSIGKeys(keys);
    BOOST_CHECK(keys.size() > 0);
    if (keys.size() > 0) {
-     BOOST_CHECK_EQUAL(keys[0].name, "test");
-     BOOST_CHECK_EQUAL(keys[0].algorithm, "NULL");
+     BOOST_CHECK_EQUAL(keys[0].name.toString(), "test.");
+     BOOST_CHECK_EQUAL(keys[0].algorithm.toString(), "NULL.");
      BOOST_CHECK_EQUAL(keys[0].key, "NULL");
    }
 }
@@ -183,8 +185,23 @@ BOOST_AUTO_TEST_CASE(test_method_setNotified) {
 BOOST_AUTO_TEST_CASE(test_method_getDomainInfo) {
    DomainInfo di;
    BOOST_TEST_MESSAGE("Testing getDomainInfo method");
-   be->getDomainInfo("unit.test", di);
-   BOOST_CHECK_EQUAL(di.zone, "unit.test");
+   be->getDomainInfo(DNSName("unit.test."), di);
+   BOOST_CHECK_EQUAL(di.zone.toString(), "unit.test.");
+   BOOST_CHECK_EQUAL(di.serial, 2);
+   BOOST_CHECK_EQUAL(di.notified_serial, 2);
+   BOOST_CHECK_EQUAL(di.kind, DomainInfo::Native);
+   BOOST_CHECK_EQUAL(di.backend, be);
+}
+
+BOOST_AUTO_TEST_CASE(test_method_getAllDomains) {
+   DomainInfo di;
+   BOOST_TEST_MESSAGE("Testing getAllDomains method");
+   vector<DomainInfo> result;
+
+   be->getAllDomains(&result, true);
+
+   di = result[0];
+   BOOST_CHECK_EQUAL(di.zone.toString(), "unit.test.");
    BOOST_CHECK_EQUAL(di.serial, 2);
    BOOST_CHECK_EQUAL(di.notified_serial, 2);
    BOOST_CHECK_EQUAL(di.kind, DomainInfo::Native);
@@ -193,8 +210,8 @@ BOOST_AUTO_TEST_CASE(test_method_getDomainInfo) {
 
 BOOST_AUTO_TEST_CASE(test_method_isMaster) {
    BOOST_TEST_MESSAGE("Testing isMaster method");
-   BOOST_CHECK(be->isMaster("ns1.unit.test", "10.0.0.1"));
-   BOOST_CHECK(!be->isMaster("ns2.unit.test", "10.0.0.2"));
+   BOOST_CHECK(be->isMaster(DNSName("ns1.unit.test."), "10.0.0.1"));
+   BOOST_CHECK(!be->isMaster(DNSName("ns2.unit.test."), "10.0.0.2"));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_superMasterBackend) {
@@ -203,20 +220,20 @@ BOOST_AUTO_TEST_CASE(test_method_superMasterBackend) {
    DNSBackend *dbd;
    BOOST_TEST_MESSAGE("Testing superMasterBackend method");
 
-   rr.qname = "example.com";
+   rr.qname = DNSName("example.com.");
    rr.qtype = QType::NS;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
-   rr.content = "ns1.example.com";
+   rr.content = "ns1.example.com.";
    nsset.push_back(rr);
-   rr.qname = "example.com";
+   rr.qname = DNSName("example.com.");
    rr.qtype = QType::NS;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
-   rr.content = "ns2.example.com";
+   rr.content = "ns2.example.com.";
    nsset.push_back(rr);
 
-   BOOST_CHECK(be->superMasterBackend("10.0.0.1", "example.com", nsset, NULL, NULL, &dbd));
+   BOOST_CHECK(be->superMasterBackend("10.0.0.1", DNSName("example.com."), nsset, NULL, NULL, &dbd));
 
    // let's see what we got
    BOOST_CHECK_EQUAL(dbd, be);
@@ -224,20 +241,20 @@ BOOST_AUTO_TEST_CASE(test_method_superMasterBackend) {
 
 BOOST_AUTO_TEST_CASE(test_method_createSlaveDomain) {
    BOOST_TEST_MESSAGE("Testing createSlaveDomain method");
-   BOOST_CHECK(be->createSlaveDomain("10.0.0.1", "pirate.unit.test", "", ""));
+   BOOST_CHECK(be->createSlaveDomain("10.0.0.1", DNSName("pirate.unit.test."), "", ""));
 }
 
 BOOST_AUTO_TEST_CASE(test_method_feedRecord) {
    DNSResourceRecord rr;
    BOOST_TEST_MESSAGE("Testing feedRecord method");
-   be->startTransaction("example.com",2);
-   rr.qname = "example.com";
+   be->startTransaction(DNSName("example.com."),2);
+   rr.qname = DNSName("example.com.");
    rr.qtype = QType::SOA;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
-   rr.content = "ns1.example.com hostmaster.example.com 2013013441 7200 3600 1209600 300";
+   rr.content = "ns1.example.com. hostmaster.example.com. 2013013441 7200 3600 1209600 300";
    BOOST_CHECK(be->feedRecord(rr, NULL));
-   rr.qname = "replace.example.com";
+   rr.qname = DNSName("replace.example.com.");
    rr.qtype = QType::A;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
@@ -247,39 +264,42 @@ BOOST_AUTO_TEST_CASE(test_method_feedRecord) {
 }
 
 BOOST_AUTO_TEST_CASE(test_method_replaceRRSet) {
-   be->startTransaction("example.com",2);
+   be->startTransaction(DNSName("example.com."),2);
    DNSResourceRecord rr;
    std::vector<DNSResourceRecord> rrset;
    BOOST_TEST_MESSAGE("Testing replaceRRSet method");
-   rr.qname = "replace.example.com";
+   rr.qname = DNSName("replace.example.com.");
    rr.qtype = QType::A;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
    rr.content = "1.1.1.1";
    rrset.push_back(rr);
-   BOOST_CHECK(be->replaceRRSet(2, "replace.example.com", QType(QType::A), rrset));
+   BOOST_CHECK(be->replaceRRSet(2, DNSName("replace.example.com."), QType(QType::A), rrset));
    be->commitTransaction();
 }
 
 BOOST_AUTO_TEST_CASE(test_method_feedEnts) {
    BOOST_TEST_MESSAGE("Testing feedEnts method");
-   be->startTransaction("example.com",2);
-   map<string, bool> nonterm = boost::assign::map_list_of("_udp", true)("_sip._udp", true);
+   be->startTransaction(DNSName("example.com."),2);
+   map<DNSName, bool> nonterm = boost::assign::map_list_of(DNSName("_udp"), true)(DNSName("_sip._udp"), true);
    BOOST_CHECK(be->feedEnts(2, nonterm));
    be->commitTransaction();
 }
 
 BOOST_AUTO_TEST_CASE(test_method_feedEnts3) {
    BOOST_TEST_MESSAGE("Testing feedEnts3 method");
-   be->startTransaction("example.com",2);
-   map<string, bool> nonterm = boost::assign::map_list_of("_udp", true)("_sip._udp", true);
-   BOOST_CHECK(be->feedEnts3(2, "example.com", nonterm, 1, "\u00aa\u00bb\u00cc\u00dd", 0));
+   be->startTransaction(DNSName("example.com"),2);
+   NSEC3PARAMRecordContent ns3prc;
+   ns3prc.d_iterations=1;
+   ns3prc.d_salt="\u00aa\u00bb\u00cc\u00dd";
+   map<DNSName, bool> nonterm = boost::assign::map_list_of(DNSName("_udp"), true)(DNSName("_sip._udp"), true);
+   BOOST_CHECK(be->feedEnts3(2, DNSName("example.com."), nonterm, ns3prc, 0));
    be->commitTransaction();
 }
 
 BOOST_AUTO_TEST_CASE(test_method_abortTransaction) {
    BOOST_TEST_MESSAGE("Testing abortTransaction method");
-   be->startTransaction("example.com",2);
+   be->startTransaction(DNSName("example.com."),2);
    BOOST_CHECK(be->abortTransaction());
 }
 
@@ -287,10 +307,15 @@ BOOST_AUTO_TEST_CASE(test_method_calculateSOASerial) {
    SOAData sd;
    time_t serial;
  
-   be->getSOA("unit.test",sd);
-   BOOST_CHECK(be->calculateSOASerial("unit.test",sd,serial));
+   be->getSOA(DNSName("unit.test."),sd);
+   BOOST_CHECK(be->calculateSOASerial(DNSName("unit.test."),sd,serial));
 
    BOOST_CHECK_EQUAL(serial, 2013060300);
 }
 
+BOOST_AUTO_TEST_CASE(test_method_directBackendCmd) {
+   BOOST_TEST_MESSAGE("Testing directBackendCmd method");
+   BOOST_CHECK_EQUAL(be->directBackendCmd("PING 1234"), "PING 1234");
+}
+
 BOOST_AUTO_TEST_SUITE_END();
