@@ -1,36 +1,39 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 #include "dnswriter.hh"
 #include "misc.hh"
 #include "dnsparser.hh"
-#include <boost/foreach.hpp>
+
 #include <limits.h>
 
-DNSPacketWriter::DNSPacketWriter(vector<uint8_t>& content, const string& qname, uint16_t  qtype, uint16_t qclass, uint8_t opcode)
+DNSPacketWriter::DNSPacketWriter(vector<uint8_t>& content, const DNSName& qname, uint16_t  qtype, uint16_t qclass, uint8_t opcode)
   : d_pos(0), d_content(content), d_qname(qname), d_canonic(false), d_lowerCase(false)
 {
   d_content.clear();
   dnsheader dnsheader;
-  
+
   memset(&dnsheader, 0, sizeof(dnsheader));
   dnsheader.id=0;
   dnsheader.qdcount=htons(1);
   dnsheader.opcode=opcode;
-  
+
   const uint8_t* ptr=(const uint8_t*)&dnsheader;
   uint32_t len=d_content.size();
   d_content.resize(len + sizeof(dnsheader));
   uint8_t* dptr=(&*d_content.begin()) + len;
-  
+
   memcpy(dptr, ptr, sizeof(dnsheader));
   d_stuff=0;
 
-  xfrLabel(qname, false);
-  
+  xfrName(qname, false);
+
   len=d_content.size();
   d_content.resize(len + d_record.size() + 4);
 
   ptr=&*d_record.begin();
   dptr=(&*d_content.begin()) + len;
-  
+
   memcpy(dptr, ptr, d_record.size());
 
   len+=d_record.size();
@@ -47,16 +50,22 @@ DNSPacketWriter::DNSPacketWriter(vector<uint8_t>& content, const string& qname,
   d_stuff=0xffff;
   d_labelmap.reserve(16);
   d_truncatemarker=d_content.size();
+  d_sor = 0;
+  d_rollbackmarker = 0;
+  d_recordttl = 0;
+  d_recordqtype = 0;
+  d_recordqclass = QClass::IN;
+  d_recordplace = DNSResourceRecord::ANSWER;
 }
 
 dnsheader* DNSPacketWriter::getHeader()
 {
-  return (dnsheader*)&*d_content.begin();
+  return reinterpret_cast<dnsheader*>(&*d_content.begin());
 }
 
-void DNSPacketWriter::startRecord(const string& name, uint16_t qtype, uint32_t ttl, uint16_t qclass, Place place, bool compress)
+void DNSPacketWriter::startRecord(const DNSName& name, uint16_t qtype, uint32_t ttl, uint16_t qclass, DNSResourceRecord::Place place, bool compress)
 {
-  if(!d_record.empty()) 
+  if(!d_record.empty())
     commit();
 
   d_recordqname=name;
@@ -68,21 +77,21 @@ void DNSPacketWriter::startRecord(const string& name, uint16_t qtype, uint32_t t
   d_stuff = 0;
   d_rollbackmarker=d_content.size();
 
-  if(compress && !d_recordqname.empty() && pdns_iequals(d_qname, d_recordqname)) {  // don't do the whole label compression thing if we *know* we can get away with "see question" - except when compressing the root
+  if(compress && d_recordqname.countLabels() && d_qname==d_recordqname) {  // don't do the whole label compression thing if we *know* we can get away with "see question" - except when compressing the root
     static unsigned char marker[2]={0xc0, 0x0c};
     d_content.insert(d_content.end(), (const char *) &marker[0], (const char *) &marker[2]);
   }
   else {
-    xfrLabel(d_recordqname, compress);
+    xfrName(d_recordqname, compress);
     d_content.insert(d_content.end(), d_record.begin(), d_record.end());
     d_record.clear();
   }
 
   d_stuff = sizeof(dnsrecordheader); // this is needed to get compressed label offsets right, the dnsrecordheader will be interspersed
-  d_sor=d_content.size() + d_stuff; // start of real record 
+  d_sor=d_content.size() + d_stuff; // start of real record
 }
 
-void DNSPacketWriter::addOpt(int udpsize, int extRCode, int Z, const vector<pair<uint16_t,string> >& options)
+void DNSPacketWriter::addOpt(uint16_t udpsize, int extRCode, int Z, const vector<pair<uint16_t,string> >& options)
 {
   uint32_t ttl=0;
 
@@ -92,16 +101,17 @@ void DNSPacketWriter::addOpt(int udpsize, int extRCode, int Z, const vector<pair
   stuff.version=0;
   stuff.Z=htons(Z);
 
+  static_assert(sizeof(EDNS0Record) == sizeof(ttl), "sizeof(EDNS0Record) must match sizeof(ttl)");
   memcpy(&ttl, &stuff, sizeof(stuff));
 
   ttl=ntohl(ttl); // will be reversed later on
-  
-  startRecord("", ns_t_opt, ttl, udpsize, ADDITIONAL, false);
+
+  startRecord(DNSName("."), QType::OPT, ttl, udpsize, DNSResourceRecord::ADDITIONAL, false);
   for(optvect_t::const_iterator iter = options.begin(); iter != options.end(); ++iter) {
     xfr16BitInt(iter->first);
     xfr16BitInt(iter->second.length());
     xfrBlob(iter->second);
-  } 
+  }
 }
 
 void DNSPacketWriter::xfr48BitInt(uint64_t val)
@@ -109,10 +119,10 @@ void DNSPacketWriter::xfr48BitInt(uint64_t val)
   unsigned char bytes[6];
   uint16_t theLeft = htons((val >> 32)&0xffffU);
   uint32_t theRight = htonl(val & 0xffffffffU);
-  memcpy(bytes, (void*)&theLeft, 2);
-  memcpy(bytes+2, (void*)&theRight, 4);
+  memcpy(bytes, (void*)&theLeft, sizeof(theLeft));
+  memcpy(bytes+2, (void*)&theRight, sizeof(theRight));
 
-  d_record.insert(d_record.end(), bytes, bytes + 6);
+  d_record.insert(d_record.end(), bytes, bytes + sizeof(bytes));
 }
 
 
@@ -136,99 +146,120 @@ void DNSPacketWriter::xfr8BitInt(uint8_t val)
 }
 
 
-/* input: 
+/* input:
+ if lenField is true
   "" -> 0
   "blah" -> 4blah
   "blah" "blah" -> output 4blah4blah
   "verylongstringlongerthan256....characters" \xffverylongstring\x23characters (autosplit)
   "blah\"blah" -> 9blah"blah
   "blah\97" -> 5blahb
+
+ if lenField is false
+  "blah" -> blah
+  "blah\"blah" -> blah"blah
   */
-void DNSPacketWriter::xfrText(const string& text, bool)
+void DNSPacketWriter::xfrText(const string& text, bool, bool lenField)
 {
   if(text.empty()) {
     d_record.push_back(0);
     return;
   }
   vector<string> segments = segmentDNSText(text);
-  BOOST_FOREACH(const string& str, segments) {
-    d_record.push_back(str.length());
+  for(const string& str :  segments) {
+    if(lenField)
+      d_record.push_back(str.length());
     d_record.insert(d_record.end(), str.c_str(), str.c_str() + str.length());
   }
 }
 
-DNSPacketWriter::lmap_t::iterator find(DNSPacketWriter::lmap_t& lmap, const string& label)
+void DNSPacketWriter::xfrUnquotedText(const string& text, bool lenField)
+{
+  if(text.empty()) {
+    d_record.push_back(0);
+    return;
+  }
+  if(lenField)
+    d_record.push_back(text.length());
+  d_record.insert(d_record.end(), text.c_str(), text.c_str() + text.length());
+}
+
+/* FIXME400: check that this beats a map */
+DNSPacketWriter::lmap_t::iterator find(DNSPacketWriter::lmap_t& nmap, const DNSName& name)
 {
   DNSPacketWriter::lmap_t::iterator ret;
-  for(ret=lmap.begin(); ret != lmap.end(); ++ret)
-    if(pdns_iequals(ret->first ,label))
+  for(ret=nmap.begin(); ret != nmap.end(); ++ret)
+    if(ret->first == name)
       break;
   return ret;
 }
 
-//! tokenize a label into parts, the parts describe a begin offset and an end offset
-bool labeltokUnescape(labelparts_t& parts, const string& label)
+// //! tokenize a label into parts, the parts describe a begin offset and an end offset
+// bool labeltokUnescape(labelparts_t& parts, const string& label)
+// {
+//   string::size_type epos = label.size(), lpos(0), pos;
+//   bool unescapedSomething = false;
+//   const char* ptr=label.c_str();
+
+//   parts.clear();
+
+//   for(pos = 0 ; pos < epos; ++pos) {
+//     if(ptr[pos]=='\\') {
+//       pos++;
+//       unescapedSomething = true;
+//       continue;
+//     }
+//     if(ptr[pos]=='.') {
+//       parts.push_back(make_pair(lpos, pos));
+//       lpos=pos+1;
+//     }
+//   }
+
+//   if(lpos < pos)
+//     parts.push_back(make_pair(lpos, pos));
+//   return unescapedSomething;
+// }
+
+// this is the absolute hottest function in the pdns recursor
+void DNSPacketWriter::xfrName(const DNSName& name, bool compress, bool)
 {
-  string::size_type epos = label.size(), lpos(0), pos;
-  bool unescapedSomething = false;
-  const char* ptr=label.c_str();
-
-  parts.clear();
-
-  for(pos = 0 ; pos < epos; ++pos) {
-    if(ptr[pos]=='\\') {
-      pos++;
-      unescapedSomething = true;
-      continue;
-    }
-    if(ptr[pos]=='.') {
-      parts.push_back(make_pair(lpos, pos));
-      lpos=pos+1;
-    }
-  }
-  
-  if(lpos < pos)
-    parts.push_back(make_pair(lpos, pos));
-  return unescapedSomething;
-}
-
-// this is the absolute hottest function in the pdns recursor 
-void DNSPacketWriter::xfrLabel(const string& Label, bool compress)
-{
-  string label = d_lowerCase ? toLower(Label) : Label;
-  labelparts_t parts;
+  //cerr<<"xfrName: name=["<<name.toString()<<"] compress="<<compress<<endl;
+  // string label = d_lowerCase ? toLower(Label) : Label;
+  // FIXME400: we ignore d_lowerCase for now
+  // cerr<<"xfrName writing ["<<name.toString()<<"]"<<endl;
+  std::vector<std::string> parts = name.getRawLabels();
+  // labelparts_t parts;
+  // cerr<<"labelcount: "<<parts.size()<<endl;
 
   if(d_canonic)
     compress=false;
 
-  string::size_type labellen = label.size();
-  if(labellen==1 && label[0]=='.') { // otherwise we encode '..'
+  if(!parts.size()) { // otherwise we encode '..'
     d_record.push_back(0);
     return;
   }
-  bool unescaped=labeltokUnescape(parts, label); 
-  
+
   // d_stuff is amount of stuff that is yet to be written out - the dnsrecordheader for example
-  unsigned int pos=d_content.size() + d_record.size() + d_stuff; 
-  string chopped;
-  bool deDot = labellen && (label[labellen-1]=='.'); // make sure we don't store trailing dots in the labelmap
+  unsigned int pos=d_content.size() + d_record.size() + d_stuff;
+  // bool deDot = labellen && (label[labellen-1]=='.'); // make sure we don't store trailing dots in the labelmap
 
   unsigned int startRecordSize=d_record.size();
   unsigned int startPos;
 
-  for(labelparts_t::const_iterator i=parts.begin(); i!=parts.end(); ++i) {
-    if(deDot)
-      chopped.assign(label.c_str() + i->first, labellen - i->first -1);
-    else
-      chopped.assign(label.c_str() + i->first);
+  DNSName towrite = name;
+  /* FIXME400: if we are not compressing, there is no reason to work per-label */
+  for(auto &label: parts) {
+    if(d_lowerCase) label=toLower(label);
+    //cerr<<"xfrName labelpart ["<<label<<"], left to write ["<<towrite.toString()<<"]"<<endl;
 
-    lmap_t::iterator li=d_labelmap.end();
+    auto li=d_labelmap.end();
     // see if we've written out this domain before
-    // cerr<<"Searching for compression pointer to '"<<chopped<<"', "<<d_labelmap.size()<<" cmp-records"<<endl;
-    if(compress && (li=find(d_labelmap, chopped))!=d_labelmap.end()) {
-      // cerr<<"\tFound a compression pointer to '"<<chopped<<"': "<<li->second<<endl;
-      if (d_record.size() - startRecordSize + chopped.size() > 253) // chopped does not include a length octet for the first label and the root label
-        throw MOADNSException("DNSPacketWriter::xfrLabel() found overly large (compressed) name");
+    //cerr<<"compress="<<compress<<", searching? for compression pointer to '"<<towrite.toString()<<"', "<<d_labelmap.size()<<" cmp-records"<<endl;
+    if(compress && (li=find(d_labelmap, towrite))!=d_labelmap.end()) {
+      //cerr<<"doing compression, my label=["<<label<<"] found match ["<<li->first.toString()<<"]"<<endl;
+      //cerr<<"\tFound a compression pointer to '"<<towrite.toString()<<"': "<<li->second<<endl;
+      if (d_record.size() - startRecordSize + label.size() > 253) // chopped does not include a length octet for the first label and the root label
+        throw MOADNSException("DNSPacketWriter::xfrName() found overly large (compressed) name");
       uint16_t offset=li->second;
       offset|=0xc000;
       d_record.push_back((char)(offset >> 8));
@@ -238,45 +269,32 @@ void DNSPacketWriter::xfrLabel(const string& Label, bool compress)
 
     if(li==d_labelmap.end() && pos< 16384) {
       //      cerr<<"\tStoring a compression pointer to '"<<chopped<<"': "<<pos<<endl;
-      d_labelmap.push_back(make_pair(chopped, pos));                       //  if untrue, we need to count - also, don't store offsets > 16384, won't work
+      d_labelmap.push_back(make_pair(towrite, pos));                       //  if untrue, we need to count - also, don't store offsets > 16384, won't work
+      //cerr<<"stored ["<<towrite.toString()<<"] at pos "<<pos<<endl;
     }
 
     startPos=pos;
 
-    if(unescaped) {
-      string part(label.c_str() + i -> first, i->second - i->first);
-
-      // FIXME: this relies on the semi-canonical escaped output from getLabelFromContent
-      boost::replace_all(part, "\\.", ".");
-      boost::replace_all(part, "\\032", " ");
-      boost::replace_all(part, "\\\\", "\\"); 
-
-      d_record.push_back(part.size());
-      unsigned int len=d_record.size();
-      d_record.resize(len + part.size());
-      memcpy(((&*d_record.begin()) + len), part.c_str(), part.size());
-      pos+=(part.size())+1;
-    }
-    else {
-      char labelsize=(char)(i->second - i->first);
-      d_record.push_back(labelsize);
-      unsigned int len=d_record.size();
-      d_record.resize(len + i->second - i->first);
-      memcpy(((&*d_record.begin()) + len), label.c_str() + i-> first, i->second - i->first);
-      pos+=(i->second - i->first)+1;
-    }
+    char labelsize=label.size();
+    //cerr<<"labelsize = "<<int(labelsize)<<" for label ["<<label<<"]"<<endl;
+    d_record.push_back(labelsize);
+    unsigned int len=d_record.size();
+    d_record.resize(len + labelsize);
+    memcpy(((&*d_record.begin()) + len), label.c_str(), labelsize); // FIXME400 do not want memcpy
+    pos+=labelsize+1;
 
     if(pos - startPos == 1)
-      throw MOADNSException("DNSPacketWriter::xfrLabel() found empty label in the middle of name");
+      throw MOADNSException("DNSPacketWriter::xfrName() found empty label in the middle of name");
     if(pos - startPos > 64)
-      throw MOADNSException("DNSPacketWriter::xfrLabel() found overly large label in name");
+      throw MOADNSException("DNSPacketWriter::xfrName() found overly large label in name");
+    towrite.chopOff();   /* FIXME400: iterating the label vector while keeping this chopoff in sync is a hack */
   }
   d_record.push_back(0); // insert root label
 
   if (d_record.size() - startRecordSize > 255)
-    throw MOADNSException("DNSPacketWriter::xfrLabel() found overly large name");
+    throw MOADNSException("DNSPacketWriter::xfrName() found overly large name");
 
-  out:;
+ out:;
 }
 
 void DNSPacketWriter::xfrBlob(const string& blob, int  )
@@ -331,7 +349,7 @@ void DNSPacketWriter::commit()
   drh.d_class=htons(d_recordqclass);
   drh.d_ttl=htonl(d_recordttl);
   drh.d_clen=htons(d_record.size());
-  
+
   // and write out the header
   const uint8_t* ptr=(const uint8_t*)&drh;
   d_content.insert(d_content.end(), ptr, ptr+sizeof(drh));
@@ -343,22 +361,19 @@ void DNSPacketWriter::commit()
 
   dnsheader* dh=reinterpret_cast<dnsheader*>( &*d_content.begin());
   switch(d_recordplace) {
-  case ANSWER:
+  case DNSResourceRecord::QUESTION:
+    dh->qdcount = htons(ntohs(dh->qdcount) + 1);
+    break;
+  case DNSResourceRecord::ANSWER:
     dh->ancount = htons(ntohs(dh->ancount) + 1);
     break;
-  case AUTHORITY:
+  case DNSResourceRecord::AUTHORITY:
     dh->nscount = htons(ntohs(dh->nscount) + 1);
     break;
-  case ADDITIONAL:
+  case DNSResourceRecord::ADDITIONAL:
     dh->arcount = htons(ntohs(dh->arcount) + 1);
     break;
   }
 
   d_record.clear();   // clear d_record, ready for next record
 }
-
-
-
-
-
-
