@@ -19,8 +19,12 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 #include "dynmessenger.hh"
 #include <cstdio>
+#include "utility.hh"
 #include <cstdlib>
 #include <cstring>
 #include <cerrno>
@@ -33,7 +37,7 @@ DynMessenger::DynMessenger(const string &fname,
     int timeout_usec)
 {
   d_s=socket(AF_UNIX,SOCK_STREAM,0);
-  Utility::setCloseOnExec(d_s);
+  setCloseOnExec(d_s);
   
   if(d_s<0) {
     throw PDNSException(string("socket")+strerror(errno));
@@ -73,7 +77,7 @@ DynMessenger::DynMessenger(const ComboAddress& remote,
     int timeout_usec)
 {
   d_s=socket(AF_INET, SOCK_STREAM,0);
-  Utility::setCloseOnExec(d_s);
+  setCloseOnExec(d_s);
  
   if(d_s<0) {
     throw PDNSException(string("socket")+strerror(errno));
@@ -115,10 +119,7 @@ DynMessenger::~DynMessenger()
 int DynMessenger::send(const string &msg) const
 {
   try {
-    if(writen2(d_s, msg+"\n") < 0) { // sue me
-      perror("sendto");
-      return -1;
-    }
+    writen2(d_s, msg+"\n");
     return 0;
   } catch(std::runtime_error& e) {
     if (errno == EAGAIN)
